from pymongo import MongoClient
from datetime import datetime
import time

from .wrappers import lookupAll, DeriveDataLookupAll
from .helpers import _RunUserDefinedFunctions_, _CollectHistory_, _CollectHistoryAgg_

## DWM on a set of contact records
def dwmAll(data, mongoDb, mongoConfig, configName, writeContactHistory=True, returnHistoryId=True, returnHistoryField='historyId', histIdField={"name": "emailAddress", "value": "emailAddress"}, udfNamespace=__name__):
    '''
        Multi-record wrapper for dwmOne

        Arguments:
        * data -- list of contact records (dictionaries)
        * mongoDb -- PyMongo MongoClient DB instance (i.e., mongoDb = MongoClient('connectionString')['dbName'])
        * mongoConfig -- dictionary of mongo collections to reference for the following: config, lookup, regex, derive, contactHistory
        * configName -- name of configuration to use
        * writeContactHistory -- bool; whether or not to write a before/after snapshot to contactHistory
    '''

    configColl = mongoDb[mongoConfig['config']]

    config = configColl.find_one({"configName": configName})

    if not config:
        raise Exception("configName '" + configName + "' not found in collection '" + mongoConfig['config']) + "'"

    for row in data:
        row, historyId = dwmOne(data=row, mongoDb=mongoDb, mongoConfig=mongoConfig, config=config, writeContactHistory=writeContactHistory, returnHistoryId=returnHistoryId, histIdField=histIdField, udfNamespace=udfNamespace)
        if returnHistoryId and writeContactHistory:
            row[returnHistoryField] = historyId

    return data

## DWM order on a single record

def dwmOne(data, mongoDb, mongoConfig, config, writeContactHistory=True, returnHistoryId=True, histIdField={"name": "emailAddress", "value": "emailAddress"}, udfNamespace=__name__):

    '''
        Wrapper for individual DWM functions

        Arguments:
        * data -- single data record to clean; key values should map to system field names (i.e., "Job Role" is keyed as "jobRole", the internal DWM name, not "C_Job_Role11", which is the Eloqua name)
        * mongoDb -- PyMongo MongoClient DB instance (i.e., mongoDb = MongoClient('connectionString')['dbName'])
        * mongoConfig -- dictionary of mongo collections to reference for the following: lookup, regex, derive, contactHistory
        * configName -- name of configuration to use
        * writeContactHistory -- bool; whether or not to write a before/after snapshot to contactHistory
    '''

    ## Setup mongo collections using mongoConfig
    lookupColl = mongoDb[mongoConfig['lookup']]
    regexColl = mongoDb[mongoConfig['regex']]
    deriveColl = mongoDb[mongoConfig['derive']]
    contactHistoryColl = mongoDb[mongoConfig['contactHistory']]

    # setup history collector
    history = {}

    # get user-defined function config
    udFun = config['userDefinedFunctions']

    ## Get runtime field configuration
    fieldConfig = config['fields']

    ## Run user-defined functions
    data, history = _RunUserDefinedFunctions_(config=config, data=data, histObj=history, position="beforeGenericValidation", namespace=udfNamespace)

    # Run generic validation lookup
    data, history = lookupAll(data=data, configFields=fieldConfig, lookupType='genericLookup', coll=lookupColl, histObj=history)

    ## Run user-defined functions
    data, history = _RunUserDefinedFunctions_(config=config, data=data, histObj=history, position="beforeGenericRegex", namespace=udfNamespace)

    # Run generic validation regex
    data, history = lookupAll(data=data, configFields=fieldConfig, lookupType='genericRegex', coll=regexColl, histObj=history)

    ## Run user-defined functions
    data, history = _RunUserDefinedFunctions_(config=config, data=data, histObj=history, position="beforeFieldSpecificValidation", namespace=udfNamespace)

    # Run field-specific validation lookup
    data, history = lookupAll(data=data, configFields=fieldConfig, lookupType='fieldSpecificLookup', coll=lookupColl, histObj=history)

    ## Run user-defined functions
    data, history = _RunUserDefinedFunctions_(config=config, data=data, histObj=history, position="beforeFieldSpecificRegex", namespace=udfNamespace)

    # Run field-specific validation regex
    data, history = lookupAll(data=data, configFields=fieldConfig, lookupType='fieldSpecificRegex', coll=regexColl, histObj=history)

    ## Run user-defined functions
    data, history = _RunUserDefinedFunctions_(config=config, data=data, histObj=history, position="beforeNormalization", namespace=udfNamespace)

    # Run normalization lookup
    data, history = lookupAll(data=data, configFields=fieldConfig, lookupType='normLookup', coll=lookupColl, histObj=history)

    ## Run user-defined functions
    data, history = _RunUserDefinedFunctions_(config=config, data=data, histObj=history, position="beforeNormalizationRegex", namespace=udfNamespace)

    # Run normalization regex
    data, history = lookupAll(data=data, configFields=fieldConfig, lookupType='normRegex', coll=regexColl, histObj=history)

    ## Run user-defined functions
    data, history = _RunUserDefinedFunctions_(config=config, data=data, histObj=history, position="beforeDeriveData", namespace=udfNamespace)

    # Fill gaps / refresh derived data
    data, history = DeriveDataLookupAll(data=data, configFields=fieldConfig, coll=deriveColl, histObj=history)

    ## Run user-defined functions
    data, history = _RunUserDefinedFunctions_(config=config, data=data, histObj=history, position="afterProcessing", namespace=udfNamespace)

    # check if need to write contact change history
    if writeContactHistory:
        history['timestamp'] = int(time.time())
        history[histIdField['name']] = data[histIdField['value']]
        historyId = contactHistoryColl.insert_one(history).inserted_id

    if writeContactHistory and returnHistoryId:
        return data, historyId
    else:
        return data, None
##



import re
import sys

## misc fcns that make everything else go smooth

def _CollectHistory_(lookupType, fromVal, toVal, using='', pattern=''):
    '''
        Create field change history object

        Arguments:
        * type --
        * fromVal --
        * toVal --
        * using --
        * pattern --
    '''

    histObj = {}

    if fromVal != toVal:
        histObj[lookupType] = {"from": fromVal, "to": toVal}

        if lookupType in ['deriveValue', 'deriveRegex', 'copyValue'] and using!='':
            histObj[lookupType]["using"] = using
        if lookupType in ['genericRegex', 'fieldSpecificRegex', 'normRegex', 'deriveRegex'] and pattern!='':
            histObj[lookupType]["pattern"] = pattern

    return histObj

def _CollectHistoryAgg_(contactHist, fieldHistObj, fieldName):
    '''
        Append field change history object into contact history object

        Arguments:
        * contactHist --
        * fieldHistObj --
        * fieldName --
    '''

    if fieldHistObj!={}:
        if fieldName not in contactHist.keys():
            contactHist[fieldName] = {}
        for lookupType in fieldHistObj.keys():
            contactHist[fieldName][lookupType] = fieldHistObj[lookupType]

    return contactHist

def _DataClean_(fieldVal):
    '''
        Apply standard cleaning to a value. Same cleaning will be applied on back-end to lookup tables

        Arguments:
        * fieldVal -- field value to clean
    '''

    fieldValNew = fieldVal

    fieldValNew = fieldValNew.upper()

    fieldValNew = fieldValNew.strip()

    fieldValNew = re.sub("[\s\n\t]+", " ", fieldValNew)

    return fieldValNew

def _RunUserDefinedFunctions_(config, data, histObj, position, namespace=__name__):
    '''
        Given a configuration, run a set of user-defined functions at various positions within DWM logic

        Arguments:
        * config -- DWM configuration object; must contain a 'userDefinedFunctions' dict with dict labels matching the 'position' argument list, and have entries keyed with 1-n function names
        * data -- dictionary of data for a given row
        * histObj -- object to which field change history (if any) should be appended
        * position -- where in the DWM process to run user-defined functions: beforeGenericValidation, beforeGenericRegex, beforeFieldSpecificValidation, beforeFieldSpecificRegex,
            beforeNormalization, beforeNormalizationRegex, beforeDeriveData, afterProcessing
    '''

    udfConfig = config['userDefinedFunctions']

    if position in udfConfig:

        posConfig = udfConfig[position]

        for udf in sorted(posConfig.keys()):

            posConfigUDF = posConfig[udf]

            data, histObj = getattr(sys.modules[namespace], posConfigUDF)(data=data, histObj=histObj)

    return data, histObj



from pymongo import MongoClient
from datetime import datetime
import re

from .helpers import _CollectHistory_, _CollectHistoryAgg_, _DataClean_

def DataLookup(fieldVal, coll, lookupType, fieldName, histObj={}):
    '''
        Lookup a value to replace the input value for a given field

        Arguments:
        * fieldVal -- field value against which to perform lookup
        * coll -- pymongo client collection where lookup documents are stored
        * lookupType -- one of ['genericLookup', 'fieldSpecificLookup', 'normLookup']
        * fieldName -- lookup field name
        * histObj -- object to which field change history (if any) should be appended
    '''

    if (lookupType=='genericLookup'):
        lookupDict = {"type": lookupType, "find": _DataClean_(fieldVal)}
    elif (lookupType in ['fieldSpecificLookup', 'normLookup']):
        lookupDict = {"type": lookupType, "fieldName": fieldName, "find": _DataClean_(fieldVal)}
    else:
        raise ValueError("Invalid type")

    fieldValNew = fieldVal

    lval = coll.find_one(lookupDict)

    if lval:
        if 'replace' in lval:
            fieldValNew = lval['replace']
        else:
            fieldValNew = ''

    change = _CollectHistory_(lookupType=lookupType, fromVal=fieldVal, toVal=fieldValNew)

    histObjUpd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change, fieldName=fieldName)

    return fieldValNew, histObjUpd


def RegexLookup(fieldVal, coll, fieldName, lookupType, histObj={}):
    '''
        Lookup and perform a regex replace

        Arguments:
        * fieldVal -- field value against which to perform regex replace
        * coll -- pymongo client collection where regex documents are stored
        * fieldName -- regex field name
        * lookupType -- one of ['fieldSpecificRegex', 'genericRegex', 'normRegex']
        * histObj -- object to which field change history (if any) should be appended
    '''

    if (lookupType=='genericRegex'):
        lookupDict = {"type": lookupType}
    elif (lookupType in ['fieldSpecificRegex', 'normRegex']):
        lookupDict = {"type": lookupType, "fieldName": fieldName}
    else:
        raise ValueError("Invalid type")

    fieldValNew = fieldVal
    pattern = ''

    reVal = coll.find(lookupDict)

    for row in reVal:

        match = re.match(row['pattern'], _DataClean_(fieldValNew), flags=re.IGNORECASE)

        if match:

            if 'replace' in row:
                fieldValNew = re.sub(row['pattern'], row['replace'], _DataClean_(fieldValNew), flags=re.IGNORECASE)
            else:
                fieldValNew = re.sub(row['pattern'], '', _DataClean_(fieldValNew), flags=re.IGNORECASE)

            pattern = row['pattern']
            break

    change = _CollectHistory_(lookupType=lookupType, fromVal=fieldVal, toVal=fieldValNew, pattern=pattern)

    histObjUpd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change, fieldName=fieldName)

    return fieldValNew, histObjUpd

def DeriveDataLookup(fieldName, coll, deriveInput, overwrite=True, fieldVal='', histObj={}):
    '''
        Derive a data value given a single derivation rule

        Arguments:
        * fieldName -- field name for which to derive a value
        * coll -- pymongo client collection where derivation lookup documents are stored
        * deriveInput -- an input dictionary with the field values to lookup
            {
                "lookupField1": "lookupVal1",
                "lookupField2": "lookupVal2"
            }
        * overwrite -- bool; whether or not to replace value if one already exists
        * fieldVal -- existing field value
        * histObj -- object to which field change history (if any) should be appended
    '''
    # test 1

    lookupVals = deriveInput

    for field in lookupVals:
        lookupVals[field] = _DataClean_(lookupVals[field])

    lookupDict = {}

    lookupDict['fieldName'] = fieldName
    lookupDict['type'] = 'deriveValue'
    lookupDict['lookupVals'] = lookupVals

    # test 2
    lval = coll.find_one(lookupDict)

    fieldValNew = fieldVal
    # test 3
    if lval and (overwrite or (fieldVal=='')):
        fieldValNew = lval['value']
    # test 4
    change = _CollectHistory_(lookupType='deriveValue', fromVal=fieldVal, toVal=fieldValNew, using=deriveInput)

    histObjUpd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change, fieldName=fieldName)

    return fieldValNew, histObjUpd

def DeriveDataCopyValue(fieldName, deriveInput, overwrite, fieldVal, histObj={}):
    '''
        Derive value by copying from another field

        Arguments:
        * fieldName -- field name for which to derive a value
        * deriveInput -- an input dictionary with the field values to lookup
            {
                "lookupField1": "lookupVal1"
            }
        * overwrite -- bool; whether or not to replace value if one already exists
        * fieldVal -- existing field value
        * histObj -- object to which field change history (if any) should be appended
    '''

    if len(deriveInput)>1:
        raise Exception("more than one field/value in deriveInput")

    fieldValNew = fieldVal

    row = list(deriveInput.keys())[0]

    if (deriveInput[row] != '' and (overwrite or (fieldVal==''))):

        fieldValNew = deriveInput[row]

    change = _CollectHistory_(lookupType='copyValue', fromVal=fieldVal, toVal=fieldValNew, using=deriveInput)

    histObjUpd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change, fieldName=fieldName)

    return fieldValNew, histObjUpd

def DeriveDataRegex(fieldName, coll, deriveInput, overwrite, fieldVal, histObj={}):
    '''
        Derive field value by performing regex search on another field

        Arguments:
        * fieldName -- field name for which to derive a value
        * coll -- pymongo client collection where derivation regex documents are stored
        * deriveInput -- an input dictionary with the field values to lookup
            {
                "lookupField1": "lookupVal1"
            }
        * overwrite -- bool; whether or not to replace value if one already exists
        * fieldVal -- existing field value
        * histObj -- object to which field change history (if any) should be appended
    '''

    if len(deriveInput)>1:
        raise Exception("more than one value in deriveInput")

    fieldValNew = fieldVal

    row = list(deriveInput.keys())[0]

    pattern = ''

    if (deriveInput[row] != '' and (overwrite or (fieldVal==''))):

        lookupDict = {}

        lookupDict['type'] = 'deriveRegex'
        lookupDict['deriveFieldName'] = row
        lookupDict['fieldName'] = fieldName

        reVal = coll.find(lookupDict)

        for lval in reVal:

            match = re.match(lval['pattern'], _DataClean_(deriveInput[row]), flags=re.IGNORECASE)

            if match:

                fieldValNew = re.sub(lval['pattern'], lval['replace'], _DataClean_(deriveInput[row]), flags=re.IGNORECASE)

                pattern = lval['pattern']
                break

    change = _CollectHistory_(lookupType='deriveRegex', fromVal=fieldVal, toVal=fieldValNew, using=deriveInput, pattern=pattern)

    histObjUpd = _CollectHistoryAgg_(contactHist=histObj, fieldHistObj=change, fieldName=fieldName)

    return fieldValNew, histObjUpd



from .cleaning import DataLookup, RegexLookup, DeriveDataLookup, DeriveDataCopyValue, DeriveDataRegex
from datetime import datetime


def lookupAll(data, configFields, lookupType, coll, histObj={}):
    '''
        Perform one type of direct data lookup or regex lookup for all relevant fields. Multi-field wrapper for DataLookup and RegexLookup

        Arguments:
        * data -- Row of data to which lookup type should be applied
        * configFields -- runtime configuration dict Fields
        * lookupType -- one of ['genericLookup', 'fieldSpecificLookup', 'normLookup', 'genericRegex', 'fieldSpecificRegex', 'normRegex']
        * coll -- pymongo client collection where lookup or regex documents are stored
        * histObj -- object to which field change history (if any) should be appended
    '''

    for field in data.keys():

        if field in configFields.keys() and data[field]!='':

            if lookupType in configFields[field]["lookup"]:

                if lookupType in ['genericLookup', 'fieldSpecificLookup', 'normLookup']:

                    fieldValNew, histObj = DataLookup(fieldVal=data[field], coll=coll, lookupType=lookupType, fieldName=field, histObj=histObj)

                elif lookupType in ['genericRegex', 'fieldSpecificRegex', 'normRegex']:

                    fieldValNew, histObj = RegexLookup(fieldVal=data[field], coll=coll, fieldName=field, lookupType=lookupType, histObj=histObj)

                data[field] = fieldValNew

    return data, histObj


def DeriveDataLookupAll(data, configFields, coll, histObj={}):
    '''
        Perform one type of data derivation for all relevant fields. Wrapper for DeriveDataLookup

        Arguments:
        * data -- Row of data to which lookup type should be applied
        * configFields -- runtime configuration dict Fields
        * coll -- pymongo client collection where lookup documents are stored
        * histObj -- object to which field change history (if any) should be appended
    '''

    for field in data.keys():

        if field in configFields.keys():

            fieldVal = data[field]

            fieldValNew = fieldVal

            for deriveSet in sorted(configFields[field]['derive'].keys()):

                deriveSetConfig = configFields[field]['derive'][deriveSet]

                if set.issubset(set(deriveSetConfig['fieldSet']), data.keys()):

                    deriveInput = {}

                    for val in deriveSetConfig['fieldSet']:
                        deriveInput[val] = data[val]

                    if deriveSetConfig['type']=='deriveValue':

                        fieldValNew, histObj = DeriveDataLookup(fieldName=field, coll=coll, deriveInput=deriveInput, overwrite=deriveSetConfig['overwrite'], fieldVal=fieldVal, histObj=histObj)
                    elif deriveSetConfig['type']=='copyValue':

                        fieldValNew, histObj = DeriveDataCopyValue(fieldName=field, deriveInput=deriveInput, overwrite=deriveSetConfig['overwrite'], fieldVal=fieldVal, histObj=histObj)

                    elif deriveSetConfig['type']=='deriveRegex':

                        fieldValNew, histObj = DeriveDataRegex(fieldName=field, coll=coll, deriveInput=deriveInput, overwrite=deriveSetConfig['overwrite'], fieldVal=fieldVal, histObj=histObj)


                if fieldValNew!=fieldVal:

                    data[field] = fieldValNew

                    break

    return data, histObj



from __future__ import print_function
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import io
import codecs
import os
import sys

import dwm
#
# def readme():
#     with open('README.md') as f:
#         return f.read()

setup(
    name='dwm',
    version=dwm.__version__,
    url='https://github.com/rh-marketingops/dwm',
    license='GNU General Public License',
    author='Jeremiah Coleman',
    tests_require=['nose', 'mongomock>=3.5.0'],
    install_requires=['pymongo>=3.2.2'],
    author_email='colemanja91@gmail.com',
    description='Best practices for marketing data quality management',
    #long_description=readme(),
    packages=['dwm'],
    include_package_data=True,
    platforms='any',
    test_suite = 'nose.collector',
    classifiers = [
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
        ],
    keywords = 'marketing automation data quality cleanse washing cleaning'
)



""" dwm (Data Washing Machine) is a collection of functions that carry out Red Hat's Marketing Operations
    best practices for maintaining data in a marketing contact database. """

from .dwm import dwmAll, _CollectHistory_, _CollectHistoryAgg_

__version__ = '0.0.1'
